﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Navigation;
using System.Windows.Threading;
using Microsoft.Win32;
using WpfTesting.Core;
using WpfTesting.HighLevel;

namespace WpfTesting
{
    public class ApplicationTesting
    {

        #region [Fields]

        private Application _app;
        private Window _mainWindow;
        private Thread _mainThread;

        private readonly ISet<Thread> _blockedThread = new HashSet<Thread>(); 

        private readonly Object _monitor = new Object();

        #endregion [/Fields]


        #region [Properties]

        public Thread AppThread
        {
            get { return _mainThread; }
        }

        public bool IsLoaded
        {
            get; 
            private set;
        }

        #endregion [/Properties]


        #region [Public methods]

        public ApplicationTesting(Type appType, Type mainWindowType)
        {
            if (!appType.IsSubclassOf(typeof(Application)))
                throw new Exception("Isn't an Application");

            if (!mainWindowType.IsSubclassOf(typeof(Window)))
                throw new Exception("Isn't a main window");

            CreateAppThread(appType, mainWindowType);
        }

        public void StartApp()
        {
            _mainThread.Start();
        }

        public void ShutDown()
        {
            Dispatcher.FromThread(_mainThread).Invoke(new Action(() => _app.Shutdown(0)));
        }

        #endregion [/Public methods]


        #region [Wrappers]

        public void Click(FrameworkElement element)
        {
            CheckElementAccess(element);

            Dispatcher.FromThread(_mainThread).Invoke(new Action(() => Common.Click(element)));
        }

        public void Click(MenuItem menuItem)
        {
            CheckElementAccess(menuItem);

            if (menuItem.Parent is MenuItem)
                Click(menuItem.Parent as MenuItem);
            else
                Click(menuItem.Parent as Menu);
        }

        public void CloseMessageBox(string title)
        {
            DialogTesting.CloseMessage(title);
        }

        public void CloseOpenFileDialog(OpenFileDialog dialog, bool fileSelected)
        {
            CheckAccess(dialog);

            DialogTesting.CloseOpenFileDialog(dialog, fileSelected);
        }

        #endregion [/Wrappers]


        #region [WpfSeeker]

        private FrameworkElement tempResult;
        public FrameworkElement GetElementByName(string name)
        {
            CheckAppAccess();

            
            Dispatcher.FromThread(_mainThread).Invoke(new Action(() =>
                                                                     {
                                                                         tempResult = 
                                                                             (FrameworkElement)
                                                                             _app.MainWindow.FindName(name);
                                                                     }));

            return tempResult;
        }

        public FrameworkElement GetElementByName(Window window, string name)
        {
            CheckElementAccess(window);

            return (FrameworkElement)window.FindName(name);
        }

        public bool IsWindowExists(string title)
        {
            return WindowsAPI.IsWindowExists(title);
        }

        #endregion [/WpfSeeker]


        #region [Private methods]

        private void CreateApp(Type appType, Type mainWindowType)
        {
            _app = (Application)Activator.CreateInstance(appType);
            ReleaseThreads();
            //_app.LoadCompleted += delegate { RaiseApplicationLoaded(); };
            _mainWindow = (Window) Activator.CreateInstance(mainWindowType);
            _mainWindow.Loaded += delegate { RaiseApplicationLoaded(); };
            _app.Run(_mainWindow);
            
        }

        private void CreateAppThread(Type appType, Type mainWindowType)
        {
            _mainThread = new Thread(() => CreateApp(appType, mainWindowType));
            _mainThread.IsBackground = false;
            _mainThread.SetApartmentState(ApartmentState.STA);
        }

        private void CheckAppAccess()
        {
            while (_app == null)
                BlockThread(Thread.CurrentThread);

            while (!IsLoaded)
                BlockThread(Thread.CurrentThread);
        }

        private void CheckWindowAccess(Window window)
        {
            CheckAppAccess();

            window.Activated += delegate { ReleaseThreads(); };
            while (!window.IsActive)
                BlockThread(Thread.CurrentThread);
        }

        private void CheckElementAccess(FrameworkElement element)
        {
            CheckAppAccess();

            element.Initialized += delegate { ReleaseThreads(); };

            while (!element.IsInitialized)
                BlockThread(Thread.CurrentThread);
        }

        private void CheckAccess(FileDialog dialog)
        {
            throw new NotImplementedException();
        }

        private void BlockThread(Thread thread)
        {
            _blockedThread.Add(thread);
            thread.Suspend();
        }

        private void ReleaseThreads()
        {
            foreach (var thread in _blockedThread)
                thread.Resume();
        }

        #endregion [/Private methods]


        #region [Events]

        public void RaiseApplicationLoaded()
        {
            IsLoaded = true;
            ReleaseThreads();
        }

        #endregion [/Events]

    }
}
